==> thx_api-0.3.9-py3-none-any.whl: thx/api/mixins.py <==
# -*- coding: utf-8 -*-

"""
Model view Mixins
============

Per rendere più strutturate le nostre View, e per renderla più omogenee sono
stati creati dei mixin con campi e metodi (vedi in particolare :class:`QuerysetMixin`),
oppure solo come superclassi per un utilizzo futuro:

* :class:`QuerysetMixin`
* :class:`ThuxStatusUpdateModelMixin`
* :class:`ThuxLightModelMixin`
* :class:`ThuxListModelMixin`
* :class:`ThuxRetrieveModelMixin`
* :class:`ThuxCreateModelMixin`
* :class:`ThuxUpdateModelMixin`
* :class:`ThuxDestroyModelMixin`
* :class:`ThuxStatusModelMixin`
* :class:`ThuxModelMixin`

Esempio
--------

::
  from thx.api.views import ThuxModelMixin

  class TemplateViewSet(ThuxModelMixin, viewsets.ModelViewSet):
      queryset = models.Template.objects.all()
      serializer_class = serializers.TemplateSerializer

      def check_permissions(self, request):
        ....

Classi
========

.. autoclass:: QuerysetMixin
   :members:

.. autoclass:: ThuxStatusUpdateModelMixin
   :members:

.. autoclass:: ThuxLightModelMixin
   :members:

.. autoclass:: ThuxListModelMixin
   :members:

.. autoclass:: ThuxRetrieveModelMixin
   :members:

.. autoclass:: ThuxCreateModelMixin
   :members:

.. autoclass:: ThuxUpdateModelMixin
   :members:

.. autoclass:: ThuxDestroyModelMixin
   :members:

.. autoclass:: ThuxStatusModelMixin
   :members:

.. autoclass:: ThuxModelMixin
   :members:

"""
from django.core.exceptions import FieldError, ValidationError
from django.http import Http404

from rest_framework.decorators import action
from rest_framework.response import Response

from . import exceptions


class QuerysetMixin:
    """
    **QuerysetMixin** è una classe che aggiunge funzionalità per manipolare i queryset impostando query_params precisi
    nella chiamata; ogni metodo gestisce una tipologia di filtri:

    * :meth:`get_filters`
    * :meth:`get_excludes`
    * :meth:`get_order_by`

    Entrambe le FK sono definite on_delete=models.PROTECT
    """
    def get_filters(self):
        """
        costruisce i filtri per il queryset leggendo il querystring e cercando il seguente pattern::

            filter__[dj_filter]=[value]

        Questa sintassi indica che il queryset dovrà essere filtrato con `dj_filter` e con valore `value`::

        /api/.../template/?filter__typology__code=flow&filter__status__lte=1

        che si traduce in questo filtro::

            filters = {
                "typology__code": "flow",
                "status__lte": 1
             }
            queryset = queryset.filter(**filters)


        può essere utilizzato più volte nella stessa url
        """
        filters = dict()
        for key, value in self.request.GET.items():
            if key.startswith('filter__'):
                key = key[8:]
                if key.endswith('isnull'):
                    if value.lower() == 'true':
                        filters[key] = True
                    elif value.lower() == 'false':
                        filters[key] = False
                    else:
                        filters[key] = value
                elif key.endswith('__in'):
                    filters[key] = value.split(",")
                else:
                    filters[key] = value
        return filters

    def get_excludes(self):
        """
        costruisce gli `exclude` per il queryset leggendo il querystring e cercando il seguente pattern::

            exclude__[dj_exclude]=[value]

        Questa sintassi indica che al queryset dovranno essere applicati i `dj_exclude` con valore `value`::

        /api/.../template/?exclude__typology__code=flow&exclude__status__lte=1

        che si traduce in questo filtro::

            queryset = queryset.exclude(typology__code="flow").exclude(status__lte=1)


        può essere utilizzato più volte nella stessa url
        """
        excludes = dict()
        for key, value in self.request.GET.items():
            if key.startswith('exclude__'):
                key = key[9:]
                if key.endswith('isnull'):
                    if value.lower() == 'true':
                        excludes[key] = True
                    elif value.lower() == 'false':
                        excludes[key] = False
                    else:
                        excludes[key] = value
                elif key.endswith('__in'):
                    excludes[key] = value.split(",")
                else:
                    excludes[key] = value
        return excludes

    def get_order_by(self):
        """
        costruisce l'order by per il queryset leggendo il querystring e cercando il seguente pattern::

            order_by=[order1.order2.-order3]

        Questa sintassi indica che il queryset dovrà essere ordinato con order1, order2 `dj_exclude`::

        /api/.../template/?order_by=order1.order2.-order3

        che si traduce in questo order_by::

            queryset = queryset.order_by("order1", "order2", "-order3")

        """
        order_by = list()
        for key, value in self.request.GET.items():
            if key.startswith('order_by'):
                order_by = value.split('.')
        return order_by

    def get_queryset(self):
        """
        utilizza in cascata i metodi get_filters(), get_excludes(), get_order_by() e li applica al queryset
        """
        filters = self.get_filters()
        excludes = self.get_excludes()
        order_by = self.get_order_by()
        queryset = super(QuerysetMixin, self).get_queryset()
        try:
            if filters:
                queryset = queryset.filter(**filters)
            if excludes:
                for key, value in excludes.items():
                    queryset = queryset.exclude(**{key: value})
            if order_by:
                queryset = queryset.order_by(*order_by)
        except ValueError:
            # just a little philosophy but it may be correct to raise a 400 bad request
            # without information
            raise Http404
        except FieldError:
            # TODO migliorare messaggio di errore
            raise exceptions.BadRequest
        except ValidationError as e:
            raise exceptions.BadRequest(detail='\n'.join(e.messages))
        return queryset.distinct()

    @property
    def paginator(self):
        """
        fa si che la response della vista non sarà paginata, leggendo il querystring e cercando il seguente pattern::

            no_page=no_page

            /api/.../template/?no_page=no_page

        """
        if self.request.query_params.get('no_page'):
            self.pagination_class = None
        return super(QuerysetMixin, self).paginator

    def paginated_list(self, request, queryset=None, serializer_class=None, *args, **kwargs):
        """
        Funzione di utilità da richiamare nelle views per ottenere una lista paginata
        evitando di riscrivere completamente la logica della list qualora cambi
        solo il queryset o il serializer di riferimento.

        Ex.
        @action(methods=['get'], detail=False, url_path='light-list')
        def light_list(self, request, *args, **kwargs):
        queryset = self.get_queryset().values(
            "pk", "project__name", "status", "date_create", "project__organization", "project__organization__name"
        ).order_by("-status", "-date_create")
        return self.paginated_list(request=request, queryset=queryset, *args, **kwargs)

        :param request: drf request
        :param queryset: non standard queryset (example .values() or .only() Or .defer())
        :param serializer_class: specific serializer class instead of default serializer
        :return: paginated Response
        """
        queryset = self.filter_queryset(queryset if queryset is not None else self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, serializer_class=serializer_class, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, serializer_class=serializer_class, many=True)
        return Response(serializer.data)
    
    def get_serializer(self, *args, **kwargs):
        serializer_class = kwargs.pop("serializer_class", None) or self.get_serializer_class()
        kwargs.setdefault('context', self.get_serializer_context())
        return serializer_class(*args, **kwargs)
        

class ThuxStatusUpdateModelMixin:
    """
    **ThuxStatusUpdateModelMixin** è un mixin che riscrive il metodo put per impostare lo status del modello con il valore
    inviato nel payload; solitamente i valori sono 1 o 0 (enabled o disabled)
    """

    def put(self, request, *args, **kwargs):
        instance = self.get_object()
        instance.status = self.new_status
        instance.save()
        serializer = self.get_serializer(instance)
        return Response(serializer.data)


class ThuxLightModelMixin:
    """
    **ThuxLightModelMixin** è un mixin che definisce il metodo `light` allo stesso modo del metodo `list` del mixin
    `ListModelMixin`. Utilizzare ereditando la classe da un viewset.
    """

    @action(methods=['get'], detail=False, url_path='light')
    def light(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)


class ThuxListModelMixin(QuerysetMixin):
    """
    **ThuxListModelMixin** è un mixin vuoto predisposto per un utilizzo futuro
    """


class ThuxRetrieveModelMixin:
    """
    **ThuxRetrieveModelMixin** è un mixin vuoto predisposto per un utilizzo futuro
    """


class ThuxCreateModelMixin:
    """
    **ThuxCreateModelMixin** è un mixin vuoto predisposto per un utilizzo futuro
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user = None


class ThuxUpdateModelMixin:
    """
    **ThuxUpdateModelMixin** è un mixin vuoto predisposto per un utilizzo futuro
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user = None


class ThuxDestroyModelMixin:
    """
    **ThuxDestroyModelMixin** è un mixin vuoto predisposto per un utilizzo futuro
    """


class ThuxStatusModelMixin:
    """
    **ThuxStatusModelMixin** è un mixin vuoto predisposto per un utilizzo futuro
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user = None


class ThuxModelMixin(
    ThuxListModelMixin, ThuxLightModelMixin, ThuxRetrieveModelMixin,
    ThuxCreateModelMixin, ThuxUpdateModelMixin, ThuxDestroyModelMixin
):
    """
    **ThuxModelMixin** è un mixin che importa  in blocco i mixin
    * :class:`ThuxListModelMixin`
    * :class:`ThuxLightModelMixin`
    * :class:`ThuxRetrieveModelMixin`
    * :class:`ThuxCreateModelMixin`
    * :class:`ThuxUpdateModelMixin`
    * :class:`ThuxDestroyModelMixin`
    * :class:`ThuxStatusModelMixin`
    """

==> thx_api-0.3.11-py3-none-any.whl: thx/api/mixins.py <==
# -*- coding: utf-8 -*-

"""
Model view Mixins
============

Per rendere più strutturate le nostre View, e per renderla più omogenee sono
stati creati dei mixin con campi e metodi (vedi in particolare :class:`QuerysetMixin`),
oppure solo come superclassi per un utilizzo futuro:

* :class:`QuerysetMixin`
* :class:`ThuxStatusUpdateModelMixin`
* :class:`ThuxLightModelMixin`
* :class:`ThuxListModelMixin`
* :class:`ThuxRetrieveModelMixin`
* :class:`ThuxCreateModelMixin`
* :class:`ThuxUpdateModelMixin`
* :class:`ThuxDestroyModelMixin`
* :class:`ThuxStatusModelMixin`
* :class:`ThuxModelMixin`

Esempio
--------

::
  from thx.api.views import ThuxModelMixin

  class TemplateViewSet(ThuxModelMixin, viewsets.ModelViewSet):
      queryset = models.Template.objects.all()
      serializer_class = serializers.TemplateSerializer

      def check_permissions(self, request):
        ....

Classi
========

.. autoclass:: QuerysetMixin
   :members:

.. autoclass:: ThuxStatusUpdateModelMixin
   :members:

.. autoclass:: ThuxLightModelMixin
   :members:

.. autoclass:: ThuxListModelMixin
   :members:

.. autoclass:: ThuxRetrieveModelMixin
   :members:

.. autoclass:: ThuxCreateModelMixin
   :members:

.. autoclass:: ThuxUpdateModelMixin
   :members:

.. autoclass:: ThuxDestroyModelMixin
   :members:

.. autoclass:: ThuxStatusModelMixin
   :members:

.. autoclass:: ThuxModelMixin
   :members:

"""
from django.core.exceptions import FieldError, ValidationError
from django.http import Http404

from rest_framework.decorators import action
from rest_framework.response import Response

from .deprecation import deprecated, RemovedInNextThxVersionWarning
from . import exceptions


paginate_list_message = "Use custom_list() instead. It has the same signature except that " \
                         "custom_qs became queryset and custom_serializer became serializer_class"


class QuerysetMixin:
    """
    **QuerysetMixin** è una classe che aggiunge funzionalità per manipolare i queryset impostando query_params precisi
    nella chiamata; ogni metodo gestisce una tipologia di filtri:

    * :meth:`get_filters`
    * :meth:`get_excludes`
    * :meth:`get_order_by`

    Entrambe le FK sono definite on_delete=models.PROTECT
    """
    def get_filters(self):
        """
        costruisce i filtri per il queryset leggendo il querystring e cercando il seguente pattern::

            filter__[dj_filter]=[value]

        Questa sintassi indica che il queryset dovrà essere filtrato con `dj_filter` e con valore `value`::

        /api/.../template/?filter__typology__code=flow&filter__status__lte=1

        che si traduce in questo filtro::

            filters = {
                "typology__code": "flow",
                "status__lte": 1
             }
            queryset = queryset.filter(**filters)


        può essere utilizzato più volte nella stessa url
        """
        filters = dict()
        for key, value in self.request.GET.items():
            if key.startswith('filter__'):
                key = key[8:]
                if key.endswith('isnull'):
                    if value.lower() == 'true':
                        filters[key] = True
                    elif value.lower() == 'false':
                        filters[key] = False
                    else:
                        filters[key] = value
                elif key.endswith('__in'):
                    filters[key] = value.split(",")
                else:
                    filters[key] = value
        return filters

    def get_excludes(self):
        """
        costruisce gli `exclude` per il queryset leggendo il querystring e cercando il seguente pattern::

            exclude__[dj_exclude]=[value]

        Questa sintassi indica che al queryset dovranno essere applicati i `dj_exclude` con valore `value`::

        /api/.../template/?exclude__typology__code=flow&exclude__status__lte=1

        che si traduce in questo filtro::

            queryset = queryset.exclude(typology__code="flow").exclude(status__lte=1)


        può essere utilizzato più volte nella stessa url
        """
        excludes = dict()
        for key, value in self.request.GET.items():
            if key.startswith('exclude__'):
                key = key[9:]
                if key.endswith('isnull'):
                    if value.lower() == 'true':
                        excludes[key] = True
                    elif value.lower() == 'false':
                        excludes[key] = False
                    else:
                        excludes[key] = value
                elif key.endswith('__in'):
                    excludes[key] = value.split(",")
                else:
                    excludes[key] = value
        return excludes

    def get_order_by(self):
        """
        costruisce l'order by per il queryset leggendo il querystring e cercando il seguente pattern::

            order_by=[order1.order2.-order3]

        Questa sintassi indica che il queryset dovrà essere ordinato con order1, order2 `dj_exclude`::

        /api/.../template/?order_by=order1.order2.-order3

        che si traduce in questo order_by::

            queryset = queryset.order_by("order1", "order2", "-order3")

        """
        order_by = list()
        for key, value in self.request.GET.items():
            if key.startswith('order_by'):
                order_by = value.split('.')
        return order_by

    def get_queryset(self):
        """
        utilizza in cascata i metodi get_filters(), get_excludes(), get_order_by() e li applica al queryset
        """
        filters = self.get_filters()
        excludes = self.get_excludes()
        order_by = self.get_order_by()
        queryset = super(QuerysetMixin, self).get_queryset()
        try:
            if filters:
                queryset = queryset.filter(**filters)
            if excludes:
                for key, value in excludes.items():
                    queryset = queryset.exclude(**{key: value})
            if order_by:
                queryset = queryset.order_by(*order_by)
        except ValueError:
            # just a little philosophy but it may be correct to raise a 400 bad request
            # without information
            raise Http404
        except FieldError:
            # TODO migliorare messaggio di errore
            raise exceptions.BadRequest
        except ValidationError as e:
            raise exceptions.BadRequest(detail='\n'.join(e.messages))
        return queryset.distinct()

    @property
    def paginator(self):
        """
        fa si che la response della vista non sarà paginata, leggendo il querystring e cercando il seguente pattern::

            no_page=no_page

            /api/.../template/?no_page=no_page

        """
        if self.request.query_params.get('no_page'):
            self.pagination_class = None
        return super(QuerysetMixin, self).paginator

    @deprecated(version=RemovedInNextThxVersionWarning.version, help_message=paginate_list_message)
    def paginated_list(self, request, queryset=None, serializer=None, *args, **kwargs):
        """
        funzione di utilità da richiamare nelle views per ottenere una lista paginata
        evitando di riscrivere completamente la logica della list qualora cambi
        solo il queryset o il serializer di riferimento.
        Rendendo esplicito il risultato in poche righe.

        Ex.
        @action(methods=['get'], detail=False, url_path='light-list')
        def light_list(self, request, *args, **kwargs):
        queryset = self.get_queryset().values(
            "pk", "project__name", "status", "date_create", "project__organization", "project__organization__name"
        ).order_by("-status", "-date_create")
        serializer = serializers.CustomSerializer(queryset, many=True)
        return super().paginated_list(request=request, custom_qs=queryset, *args, **kwargs)

        :param request: request obj
        :param queryset: standard or non standard queryset (example .values() or .only() Or .defer())
        :param serializer: specific serializer class instead of default serializer
        :return: paginated Response
        """
        queryset = self.filter_queryset(queryset if queryset is not None else self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = serializer if serializer else self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = serializer if serializer else self.get_serializer(queryset, many=True)
        return Response(serializer.data)

    def custom_list(self, request, queryset=None, serializer_class=None, *args, **kwargs):
        """
        Funzione di utilità da richiamare nelle views per ottenere una lista
        evitando di riscrivere completamente la logica della list qualora cambi
        solo il queryset o il serializer di riferimento.
        Ad esempio api diverse (quindi nomi diversi) che modificano il queryset ma utilizzano lo stesso serializer

        Ex.
        @action(methods=['get'], detail=False, url_path='light-list')
        def light_list(self, request, *args, **kwargs):
        queryset = self.get_queryset().values(
            "pk", "project__name", "status", "date_create", "project__organization", "project__organization__name"
        ).order_by("-status", "-date_create")
        serializer_class = serializers.CustomSerializer*
        return self.custom_list(request=request, queryset=queryset, serializer_class=serializer_class, *args, **kwargs)

        *("CustomSerializer nome di fantasia per rendere esplicativo l'esempio di norma
        dovrebbe essere NomeModelloLightListSerializer")

        :param request: drf request
        :param queryset: queryset o non standard queryset (example .values() or .only() Or .defer())
        :param serializer_class: specific serializer class instead of default serializer
        :return: paginated Response
        """
        queryset = self.filter_queryset(queryset if queryset is not None else self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, serializer_class=serializer_class, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, serializer_class=serializer_class, many=True)
        return Response(serializer.data)

    def get_serializer(self, *args, **kwargs):
        serializer_class = kwargs.pop("serializer_class", None) or self.get_serializer_class()
        kwargs.setdefault('context', self.get_serializer_context())
        return serializer_class(*args, **kwargs)


class ThuxStatusUpdateModelMixin:
    """
    **ThuxStatusUpdateModelMixin** è un mixin che riscrive il metodo put per impostare lo status del modello con il valore
    inviato nel payload; solitamente i valori sono 1 o 0 (enabled o disabled)
    """

    def put(self, request, *args, **kwargs):
        instance = self.get_object()
        instance.status = self.new_status
        instance.save()
        serializer = self.get_serializer(instance)
        return Response(serializer.data)


class ThuxLightModelMixin:
    """
    **ThuxLightModelMixin** è un mixin che definisce il metodo `light` allo stesso modo del metodo `list` del mixin
    `ListModelMixin`. Utilizzare ereditando la classe da un viewset.
    """

    @action(methods=['get'], detail=False, url_path='light')
    def light(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)


class ThuxListModelMixin(QuerysetMixin):
    """
    **ThuxListModelMixin** è un mixin vuoto predisposto per un utilizzo futuro
    """


class ThuxRetrieveModelMixin:
    """
    **ThuxRetrieveModelMixin** è un mixin vuoto predisposto per un utilizzo futuro
    """


class ThuxCreateModelMixin:
    """
    **ThuxCreateModelMixin** è un mixin vuoto predisposto per un utilizzo futuro
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user = None


class ThuxUpdateModelMixin:
    """
    **ThuxUpdateModelMixin** è un mixin vuoto predisposto per un utilizzo futuro
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user = None


class ThuxDestroyModelMixin:
    """
    **ThuxDestroyModelMixin** è un mixin vuoto predisposto per un utilizzo futuro
    """


class ThuxStatusModelMixin:
    """
    **ThuxStatusModelMixin** è un mixin vuoto predisposto per un utilizzo futuro
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user = None


class ThuxModelMixin(
    ThuxListModelMixin, ThuxLightModelMixin, ThuxRetrieveModelMixin,
    ThuxCreateModelMixin, ThuxUpdateModelMixin, ThuxDestroyModelMixin
):
    """
    **ThuxModelMixin** è un mixin che importa  in blocco i mixin
    * :class:`ThuxListModelMixin`
    * :class:`ThuxLightModelMixin`
    * :class:`ThuxRetrieveModelMixin`
    * :class:`ThuxCreateModelMixin`
    * :class:`ThuxUpdateModelMixin`
    * :class:`ThuxDestroyModelMixin`
    * :class:`ThuxStatusModelMixin`
    """
